import 'package:async_events/async_events.dart';
import 'package:logging/logging.dart' as logging;
import 'package:reflection_factory/reflection_factory.dart';
import 'package:test/test.dart';

const _epochPeriod = Duration(seconds: 2);

void main() {
  logging.Logger.root.level = logging.Level.ALL;
  logging.Logger.root.onRecord
      .listen((event) => print('${DateTime.now()}\t$event'));

  final log = logging.Logger('test');

  group('AsyncEvent', () {
    test('json', () async {
      var now = DateTime.now();

      var event1 = AsyncEvent('foo', AsyncEventID(1, 2), now, 'test', {'a': 1});
      expect(event1.toJson(), equals(event1.toJson()));

      expect(AsyncEvent.fromJson(event1.toJson()), equals(event1));

      var classReflection =
          ReflectionFactory().getRegisterClassReflection<AsyncEvent>()!;

      expect(classReflection.fromJson(event1.toJson()), equals(event1));
    });

    test('basic', () async {
      var storage =
          AsyncEventStorageMemory('test-memory', epochPeriod: _epochPeriod);
      var hub = AsyncEventHub('test', storage);

      await _doTestBasic(hub, log, storage);
    });

    test('remote', () async {
      var server = _MyAsyncEventStorageServer(
          AsyncEventStorageMemory('test-server', epochPeriod: _epochPeriod));
      var client = _MyAsyncEventStorageClient(server);

      var storage = AsyncEventStorageRemote('test', client);
      var hub = AsyncEventHub('test', storage);

      await _doTestBasic(hub, log, storage);
    });
  });
}

Future<void> _doTestBasic(
    AsyncEventHub hub, logging.Logger log, AsyncEventStorage storage) async {
  log.info('<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<');
  log.info("Hub: $hub");
  log.info("Storage: $storage");

  var c1 = hub.channel('c1');
  var c2 = hub.channel('c2');

  var c1Events = <AsyncEvent>[];
  var c2Events = <AsyncEvent>[];

  var sub1 = await c1.subscribe((e) {
    c1Events.add(e);
    log.info('C1>> $e');
  });

  var sub2 = await c2.subscribe((e) {
    c2Events.add(e);
    log.info('C2>> $e');
  });

  var eventC1_1 = await c1.submit('t', {'name': 't1'});
  var eventC2_1 = await c2.submit('t', {'name': 't2'});

  await sub1.pull();
  await sub2.pull();

  expect(c1Events.map((e) => e.toJson()), [eventC1_1!.toJson()]);
  expect(c2Events.map((e) => e.toJson()), [eventC2_1!.toJson()]);

  var c2bEvents = <AsyncEvent>[];

  var sub2bAsync = c2.subscribe(fromBegin: true, (e) {
    c2bEvents.add(e);
    log.info('C2/b>> $e');
  });

  var eventC2_2 = await c2.submit('t', {'name': 't3'});

  var sub2b = await sub2bAsync;

  await sub2b.pull();

  expect(c2bEvents.map((e) => e.toJson()),
      [eventC2_1.toJson(), eventC2_2!.toJson()]);

  expect(c2Events.map((e) => e.toJson()),
      [eventC2_1.toJson(), eventC2_2.toJson()]);

  var eventPulling = AsyncEventPulling(c1, period: Duration(milliseconds: 200));

  expect(eventPulling.period.inMilliseconds, equals(200));
  expect(eventPulling.minInterval.inMilliseconds, equals(20));
  expect(eventPulling.isStarted, isFalse);
  expect(eventPulling.isStopped, isFalse);
  expect(eventPulling.isScheduled, isFalse);

  eventPulling.start();
  expect(eventPulling.isStarted, isTrue);
  expect(eventPulling.isStopped, isFalse);

  var eventC1_2 = await c1.submit('t', {'name': 't4'});

  log.info("waitPulling 1> $eventPulling");
  expect(await eventPulling.waitPulling(), isTrue);

  expect(c1Events.map((e) => e.toJson()),
      [eventC1_1.toJson(), eventC1_2!.toJson()]);

  expect(await hub.epoch, equals(0));

  await _sleep(log, 2100);

  expect(await hub.epoch, equals(1));

  var eventC1_3 = await c1.submit('t', {'name': 't5'});

  log.info("waitPulling 2> $eventPulling");
  expect(await eventPulling.waitPulling(), isTrue);

  expect(c1Events.map((e) => e.toJson()),
      [eventC1_1.toJson(), eventC1_2.toJson(), eventC1_3!.toJson()]);

  eventPulling.stop();
  await _sleep(log, 300);

  expect(eventPulling.isStarted, isFalse);
  expect(eventPulling.isStopped, isTrue);

  expect(sub1.isSubscribed, isTrue);
  sub1.cancel();
  expect(sub1.isSubscribed, isFalse);

  expect(sub2.isSubscribed, isTrue);
  sub2.cancel();
  expect(sub2.isSubscribed, isFalse);

  expect(sub2b.isSubscribed, isTrue);
  sub2b.cancel();
  expect(sub1.isSubscribed, isFalse);

  expect(await storage.last('c1'), eventC1_3);
  expect(await storage.last('c2'), eventC2_2);

  log.info("Hub: $hub");
  log.info("Storage: $storage");
  log.info('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');
}

Future<dynamic> _sleep(logging.Logger log, int milliseconds) {
  log.info("Sleep: ${milliseconds}ms ...");
  return Future.delayed(Duration(milliseconds: milliseconds));
}

class _MyAsyncEventStorageClient extends AsyncEventStorageClient
    with AsyncEventStorageFromJSON {
  final _MyAsyncEventStorageServer server;

  @override
  AsyncEventStorageAsJSON get storageAsJSON => server;

  _MyAsyncEventStorageClient(this.server);

  @override
  String toString() {
    return '_MyAsyncEventStorageClient@$server';
  }
}

class _MyAsyncEventStorageServer with AsyncEventStorageAsJSON {
  @override
  final AsyncEventStorage storage;

  _MyAsyncEventStorageServer(this.storage);

  @override
  Future<int> get epoch => super.epoch.asFuture;

  @override
  Future<Map<String, dynamic>?> newEvent(
          String channelName, String type, Map<String, dynamic> payload) =>
      super.newEvent(channelName, type, payload).asFuture;

  @override
  Future<List<Map<String, dynamic>>> fetch(
          String channelName, AsyncEventID fromID) =>
      super.fetch(channelName, fromID).asFuture;

  @override
  Future<Map<String, dynamic>?> lastID(String channelName) =>
      super.lastID(channelName).asFuture;

  @override
  Future<Map<String, dynamic>?> last(String channelName) =>
      super.last(channelName).asFuture;

  @override
  Future<int> purge(int untilEpoch) => super.purge(untilEpoch).asFuture;

  @override
  String toString() {
    return '_MyAsyncEventStorageServer@$storage';
  }
}

extension _FutureOrExtension<T> on FutureOr<T> {
  Future<T> get asFuture {
    var self = this;
    if (self is Future<T>) {
      return self;
    } else {
      return Future.delayed(Duration(milliseconds: 10), () => self);
    }
  }
}
